#!/usr/bin/perl
#
# $RCSfile: ptksdict.plx,v $
# $Author: swaj $
# $Revision: 1.39 $
#
# Perl/Tk Sdictionary.
#
# Copyright (c) Alexey Semenoff 2001-2007. All rights reserved.
# Distributed under GNU Public License.
#
#

use 5.008;
use strict;
use warnings;
use Encode qw /encode decode from_to/;
use Getopt::Long;
use Time::HiRes qw(gettimeofday tv_interval);

use Tk;
use Tk::LabFrame;
use Tk::ProgressBar;
use Tk::ROText;

warn 'Tested with Tk 804.027 only!' if ($Tk::VERSION < 804.027);

use vars qw /

    $t0
    $adaptive_keyb
    $debug
    $data
    $text
    $data1
    $data2

    $temp
    $geometry

    $VERSION
    $NAME
    $PACKAGE

    $dictionary
    $dictionary_tmp

    %encode_converts
    %encode_autoswitch
    $encode_convert
    $sd
    $toplevel
    $toplevel_info
    $toplevel_info_t
    $toplevel_hist
    @bold
    @normal
    %config
    %DICTS

    $ControlPressed
    $AltPressed

    $mw
    $pw
    $pwl
    $pwr
    $menu
    $logo
    $img
    $ff
    $ft
    $ftb
    $fh
    $opt
    $opt2
    $fr2
    $fr2up
    $fr2down
    $fr3
    $fr4
    $fr4_lab
    $entry
    $entry_text
    $list
    $list_h
    $text
    $lscroll

    @curListView
    $curWord
    $curUp
    $curBot

    $search_entry
    $encode_convert_search

    $rb1_value
    $rb2_value

    $searchSavePos
    $searchCurPos
    $search_status
    $curWordCnt
    $stop_search
    $cleanFirstTime
    $slab
    $start_from

    $OS

    $ignoresindex
    $help
    @HISTORY
    @HISTORY_BACK
    $history_back_upd

    @HIT_CACHE
    $prev_len
    /;


BEGIN {
    $_=$0;
    s|^(.+)/.*|$1|;
    $start_from = $_;
    push @INC, ($_, "$_/lib", "$_/../lib", "$_/..", "/opt/sdict/lib" ) ;
    $toplevel_info_t = 0;

    @bold  = (
	      -foreground => '#0000ff',
	      -background => '#ffffff',
	      qw/-relief raised -borderwidth 1/,
	      );

    @normal = (
	       -foreground => undef,
	       -background => undef,
	       qw/-relief flat/,
	       );

    $ControlPressed = 0;
    @curListView = ();
    $curUp = 0;
    $curBot = 0;
    $debug = 0;
    $cleanFirstTime = 0;
    $adaptive_keyb = 0;
    $PACKAGE = __PACKAGE__;

    if ( $^O =~ /^(ms)?(dos|win(32|nt)?)/i ) {
	$OS = 'WIN';
    }
    else {
	$OS = 'UNIX';
    }

    $t0 = [gettimeofday];
    @HISTORY = ();
    @HISTORY_BACK = ();
    $history_back_upd = 1;
    @HIT_CACHE = ();
    $prev_len = 0;
    $dictionary_tmp = q{};
};

use Sdict;

use constant {

    SDICT_PROGRAM_VER      => '1.2.3'               ,
    #
    SDICT_CONFIG_FILE      => '.ptksdictrc'         ,
    SDICT_CONFIG_FILE_WIN  => 'ptksdictrc-win'      ,
    SDICT_LASTDIC_FILE     => '.ptksdictrc-last'    ,
    SDICT_LASTDIC_FILE_WIN => 'ptksdictrc-last'     ,
    SDICT_PROGRAM_TITLE    => 'Perl/Tk Sdictionary' ,
    # Engine
    SDICT_LOAD_ITEMS       => 200                   ,
    SDICT_SCROLL_MAGIC     => 0.05                  ,
    SDICT_KEYPRESS_SENS    => 0.5                   ,
    SDICT_HISTORY_LEN      => 20                    ,
    SDICT_HISTORY_BACK_LEN => 100                   ,
    # GUI
    SDICT_ENTRY_WIDTH      => 5			    ,
    SDICT_SCROLLBAR_WIDTH  => 8                     ,
    SDICT_LB_WIDTH	   => 16		    ,
};


$NAME = SDICT_PROGRAM_TITLE;
$VERSION = SDICT_PROGRAM_VER;

binmode STDOUT, ':utf8';

sub printd   (;@);
sub s_print  (;@);
sub s_printb (;@);

####
#  #
####

GetOptions(
	   "debug"         => \$debug,
	   "adaptivekeys!" => \$adaptive_keyb,
	   "ignoresindex!" => \$ignoresindex,
	   "geometry=s"	   => \$geometry,
	   "help"          => \$help,
	   );

help() if ($help);

$sd = Sdict->new;

ParseConfig();

$sd->debug_on if $debug;

DrawInterface();

s_print SDICT_PROGRAM_TITLE, 'ver.', SDICT_PROGRAM_VER, 'started';

unless ( GetAvailableDicts() ) {
    prerror 'No dictionaries found!';
    s_print 'No dictionaries found!';
}
else {
    my @dicts = sort ( keys ( %DICTS ) );

    for ( @dicts ) {
      $_ = decode ("utf8", $_);
    }

    $opt->addOptions ( @dicts );

    for ( keys %DICTS ) {
	if ( $DICTS{ $_ } eq $dictionary_tmp ) {

	  my $dctt = $_;
	  printd "using the last dct '$dctt'";
	  s_print "loading last dictionary '$dctt'";
	  $dctt = decode ( "utf8", $dctt );
	  $dictionary = $dctt;
	  $cleanFirstTime = 0;
	  ChangeDic();
	  last;
	}
    }

    @dicts = ();
}

s_print 'init OK';

MainLoop;

####
#  #
####


sub help {
     print <<EOS;

Usage: $0

    [--adaptivekeys] Introduce some delay before search, can help on slow machines
    [--ignoresindex] Ignore short index in search, use for old buggy formatted dictionaries only
    [--geometry=...] Specify geometry, (i.e --geometry=800x600+0+0)
    [--debug]        Enable debug messages to STDOUT
    [--help]         This help page

EOS
    exit;
}


sub printd (;@) {
    $debug && print "DEBUG ($PACKAGE): ", @_, "\n";
}


sub s_print (;@) {
    if ($text) {
	my $tx = join (q{ }, @_) . "\n";
	$tx  = decode ( "utf8", $tx );
	$text->insert ( 'end', $tx ) if ( $text );
	$mw->update if ( $mw );
    }
}


sub s_printb (;@) {
    if ( $text ) {
	my $tx = join (q{ }, @_) . "\n";
	$tx = decode ("utf8", $tx);
	$text->insert ('end', $tx,  ['bold']) if ( $text );
	$mw->update if ( $mw );
    }
}


sub CleanQuit {
    warn 'Quitting';
    StoreLastDic();
    $mw->destroy if $mw;
    exit;
}


sub GetAvailableDicts {
    my $patt = "$config{ dictionary_dir }/*" . Sdict::SDICT_FILE_EXT;
    my $j;
    $sd = Sdict->new;

    s_print "searching dictionaries $patt";
    prinfo "searching dictionaries '$patt'";

    my @files = glob ( $patt );
    unless ( @files ) {
	prerror 'No dictionaries found';
	return 0;
    }

    for $j (sort @files) {
	prinfo "Trying '$j'";
	$sd->init ( { file => $j } );
	unless ( $sd->read_header ) {
	    prerror "Unable to load dictionary from file '$j'";
	    next;
	}
	s_print "found '$sd->{ header }->{ title }'";
	$DICTS{ $sd->{ header }->{ title } } = $j; 
	$sd->unload_dictionary;
    }

    printd "***";

    my $lastfile = RestoreLastDic();

    unless ( $lastfile ) {
	printd "no last file";
	return 1;
    }

    for ( keys %DICTS ) {
	if ( $DICTS{ $_ } eq $lastfile ) {
	    printd "last dic found ($lastfile), setting as a default";
	    $dictionary_tmp = $lastfile;
	    return 1;
	}
    }


    printd "not found, trying to add '$lastfile'";
    $sd->init ( { file => $lastfile } );

    unless ( $sd->read_header ) {
	prerror "Unable to load dictionary from lastfile '$lastfile'";
	return 1;
    }

    s_print "found '$sd->{ header }->{ title }'";
    $DICTS{$sd->{ header }->{ title } } = $lastfile; 
    $dictionary_tmp = $lastfile;
    $sd->unload_dictionary;
    return 1;
}


sub StoreLastDic {
    my $file;

    return unless ( $config{ 'store_last_dic' } );

    if ( $OS eq 'WIN' ) {
	$file = $start_from . '/../etc/' . SDICT_LASTDIC_FILE_WIN;
    }
    else { 
	$file = "$ENV{'HOME'}/" . SDICT_LASTDIC_FILE;
    }

    unless ( $sd->{ init } ) {
	printd "No dictionary loaded, erase lastdic file";
	unlink ( $file );
	return;
    }
    else {
	my $dicc = encode( "utf8", $dictionary );
	my $lastfile = $DICTS{ $dicc };

	if ( open ( F, "> $file" ) ) {
	    print F "$lastfile\n";
	    close F;
	}
	else {
	    printd "unable to save lastdic ($file): $!";
	}
    }
}


sub RestoreLastDic {
    my $file;
    my $lastfile =q{};

    return q{} unless ( $config { 'store_last_dic' } );

    if ($OS eq 'WIN') {
	$file = $start_from . '/../etc/' . SDICT_LASTDIC_FILE_WIN;
    }
    else { 
	$file = "$ENV{'HOME'}/" . SDICT_LASTDIC_FILE;
    }

    if ( open ( F, "< $file" ) ) {
	$lastfile = <F>;
	close F;
	chomp ( $lastfile );
    }

    if ( $lastfile ) {
	printd "lastfile = '$lastfile'";
    }

    return $lastfile;
}


sub ParseConfig
{
    my $file;
    my %tempcfg = ();
    my @temp = ();
    my @temp2 = ();

    if ( $OS eq 'WIN' ) {
	$file = $start_from . '/../etc/' . SDICT_CONFIG_FILE_WIN;
    }
    else { 
	$file = "$ENV{'HOME'}/" . SDICT_CONFIG_FILE;
    }

    if ( open (F, "< $file" ) ) {
	while (<F>) {
	    chomp;
	    s/\r$//g;
	    next if /^\#/;
	    next if /^\s*$/;
	    my ( $p,$v ) = ( split ( /\s=\s/, $_, 2 ) )[0,1];
	    $p =~ s|^\s+||g;
	    $p =~ s|\s+$||g;
	    $v =~ s|^\s+||g;
	    $v =~ s|\s+$||g;
	    next unless ( $p && $v );
	    printd "config: $p -- $v";
	    $tempcfg{$p} = $v;
	}
	close F;
    }
    else {
	prinfo "Unable read configuration from file '$file'";
    }

    if ($OS eq 'WIN') {
	$config{ 'dictionary_dir' } =  $tempcfg{ 'dictionary_dir' } || $start_from . '/../dicts';
    }
    else {
	$config{ 'dictionary_dir' } = $tempcfg{ 'dictionary_dir' } || '.';
    }

    @temp = split /\s+/, $tempcfg{ 'input_encodings' }; 

    if ( @temp )
    {
	$config{ 'input_encodings' } = \@temp;
	$config{ 'disable_switcher' } = 0 ;
    } else {
	$config{ 'input_encodings' } = [] ;
	$config{ 'disable_switcher' } = 1 ;
    }

    @temp2 = split /\s+/, $tempcfg{ 'autoswitch_encoding' };
    $config{ 'autoswitch_encoding' } = @temp2 ? \@temp2 : [];

    $config{ 'font' } = $tempcfg{ 'font' } || '-*-*-*-*-*--*-*-*-*-*-*-iso10646-*';

    unless ( exists ( $tempcfg{ 'font_bold' } ) ) {
	$config{ 'font_bold' } = $config{ 'font' };
	$config{ 'font_bold' } =~ s|^(\-.+?\-.+?)\-.+?\-(.*)$|$1-bold-$2|;
    }
    else {
	$config{ 'font_bold' } = $tempcfg{ 'font_bold' };
    }

    %encode_converts   = @{ $config{ 'input_encodings'     } };
    %encode_autoswitch = @{ $config{ 'autoswitch_encoding' } };

    for ( keys %tempcfg ) {
	$config{ alt_bindings }->{ $_ } = $tempcfg{ $_ } if ( /Alt\-\d/ );
    }

    if ( exists ( $tempcfg{ 'cache_search_results' } ) ) {
	$config{ 'cache_search_results' } = 1;
    }
    else {
	$config{ 'cache_search_results' } = 0;
    }

    if ( exists ( $tempcfg{ 'store_last_dic' } ) ) {
	$config{ 'store_last_dic' } = 1;
    }
    else {
	$config{ 'store_last_dic' } = 0;
    }

    if ( exists ( $tempcfg{'format_article' } ) ) {
	$config{ 'format_article' } = 1;
    }
    else {
	$config{ 'format_article' } = 0;
    }

    if (exists ($tempcfg{ 'output_encoding' }) ) {
	$config{ 'output_encoding' } = $tempcfg{ 'output_encoding' };
    }

    return 1;
}


sub DrawInterface
{
    $mw = MainWindow->new;
    $mw->title ( $NAME );
    $mw->CmdLine;
    $mw->protocol (
		   'WM_DELETE_WINDOW' => [ \&CleanQuit ],
		   );

    $menu = $mw->Frame (
			-relief => 'raised',
			-borderwidth => 1,
			) -> pack (
				   -fill =>'x'
				   );

    $logo .= $_ while ( <DATA> );

    $img = $menu->Photo(
			-format => 'xpm',
			-data  => $logo
			);

    $debug || $mw->Icon(
			-image => $img
			);

    $pw = $mw->Panedwindow;
    $pw->pack(qw/-side top -expand yes -fill both -pady 0 -padx 0/);

    $pwl = $mw->Frame (
		       -relief => 'sunken', #'raised',
		       -borderwidth => 0,
		       ) -> pack (
				  -side => 'left',
				  -fill => 'both',
				  -expand => 'yes',
				  );

    $pwr = $mw->Frame (
		       -relief => 'sunken' ,
		       -borderwidth => 0,
		       ) -> pack (
				  -side => 'right',
				  -fill => 'both',
				  -expand => 'yes',
				  );

    $pw->add($pwl, $pwr);

    $ff = $menu->Menubutton (
			     -text => 'File', 
                             -font => $config { 'font_bold' },
                             -tearoff => 0,
                             -padx => 1,
                             -pady => 1,
                             -underline => 0,
			     )-> pack(
				      -side => 'left'
				      );
    $ff->command (
		  -label => 'Open',
		  -font => $config{ 'font_bold' },
		  -command => [ \&FileOpen ],
		  -underline => 0
		  );

    $ff->command (
		  -label => 'Info',
		  -font => $config{ 'font_bold' },
		  -command => [ \&DictionaryInfo ],
		  -underline => 0,
		  );

    $ff->command (
		  -label => 'Quit',
		  -font => $config{ 'font_bold' },
		  -command => [ \&CleanQuit ],
		  -underline => 0,
		  );

    $ft = $menu->Menubutton (
			     -text => 'Tools', 
                             -font => $config{ 'font_bold' },
                             -tearoff => 0,
                             -padx => 10,
                             -pady => 1,
                             -underline => 0,
			     )-> pack(
				      -side => 'left'
				      );


    $ftb = $menu->Menubutton (
			     -text => 'Back', 
                             -font => $config{ 'font_bold' },
                             -tearoff => 0,
                             -padx => 10,
                             -pady => 1,
                             -underline => 0,
			     )-> pack (
				       -side => 'left',
				       );

    $ftb->bind (
		'<ButtonPress-1>' => [ \&BackButton ]
		);


    $ft->command (
		  -label => 'Search',
		  -font => $config{ 'font_bold' },
		  -command => [ \&FindInArticles ],
		  -underline => 1
		  );

    $ft->command (
		  -label => 'History',
		  -font => $config{ 'font_bold' },
		  -command => [ \&WordHistory ],
		  -underline => 1
		  );

    $ft->command (
		  -label => 'Save article...',
		  -font => $config{ 'font_bold' },
		  -command => [ \&SaveArticleAsText ],
		  -underline => 1
		  );

    $fh = $menu->Button (
			 -text => 'Help',
			 -font => $config{ 'font_bold' },
			 -borderwidth => 0,
			 -padx => 1,
			 -pady => 1,
			 -command => [ \&PrintHelp ],
			 -underline => 0,
			 ) -> pack (
				    -side => 'right',
				    );

    $opt = $menu->Optionmenu(
			     -options => [ () ] ,
			     -variable => \$dictionary,
			     -command => [ \&ChangeDic ],
			     ) -> pack ();
    # status bar
    $fr4 = $mw->Frame (
		       -relief => 'sunken', #'raised',
		       -borderwidth => 1,
		       ) -> pack (
				  -side => 'bottom',
				  -fill => 'x',
				  -expand => 'no'
				  );

    $fr4_lab = $fr4->Label (
			    -text => q{ },
			    -font => $config{ 'font' },
			    -justify => 'left' ) -> pack (
							 -side => 'left', 
							  );
    $fr2 = $pwl->Frame (
		       -relief => 'raised',
		       -borderwidth => 0,
		       ) -> pack (
				  -side => 'left',
				  -fill => 'both',
				  -expand => 'yes',
				  );
    # for entry
    $fr2up = $fr2->Frame (
			  -relief => 'raised',
			  -borderwidth => 1,
			  ) -> pack (
				     -side => 'top',
				     -fill => 'x',
				     );
    # for LB
    $fr2down = $fr2->Frame (
			    -relief => 'raised',
			    -borderwidth => 1,
			    ) -> pack (
				       -side => 'bottom',
				       -fill => 'both',
				       -expand => 'yes',
				       );

    # for text
    $fr3 = $pwr->Frame (
		       -relief => 'solid',
			-borderwidth => 0,
		       ) -> pack (
				  -side => 'right',
				  -fill => 'both',
				  -expand => 'yes',
				  -padx => 0,
				  -pady => 0,
				  
				  );

    $entry = $fr2up->Entry(
			   -width => SDICT_ENTRY_WIDTH, #  30
			   -font => $config{ 'font' },
			   -textvariable => \$entry_text,
			   )-> pack (
				     -side =>'left',
				     -fill => 'x',
				     -expand => 'yes',
				     );
    $entry->bind (
		  '<Return>' => [ \&ReturnPress ]
		  );

    $entry->bind (
		  '<Key>' => [ \&EntryPress ]
		  );


    if ( $OS eq 'UNIX' ) {
	$entry->bind (
		      '<ButtonRelease-2>' => [ \&EntryPaste ]
		      );
    }
    else {
	$entry->bind (
		      '<Control-v>' => [ \&EntryPaste ]
		      );
    }

    $entry->bind (
		  '<Escape>' => sub { 
		         CleanCacheHistory();
			 $entry -> delete( 0.1, 'end' );
		     }
		  );

    if ( $OS eq 'UNIX' && ! $config{ 'disable_switcher' } ) {
	$opt2 = $fr2up->Optionmenu(
				   -options => [ () ] ,
				   -variable => \$encode_convert,
				   -command => [ \&ChangeEncoding ]
				   ) -> pack();

	$opt2->addOptions (
			   sort keys ( %encode_converts )
			   );
    }

    $list = $fr2down->Listbox(
			      -width => SDICT_LB_WIDTH,
			      );

    $list->configure (
		      -font => $config{ 'font' },
		      );

    $list->pack (
		 qw/ -side left -expand yes -fill both /
		 );

    $lscroll = $fr2down->Scrollbar(
				   orient  => 'vertical',
				   width   => SDICT_SCROLLBAR_WIDTH,
				   command => \&ScrollCmd ,
				   jump => 1
				   ) ->pack(
					    qw/ -side right -fill both /
					    );
    $lscroll->set( 0, SDICT_SCROLL_MAGIC );

    $text = $fr3->Scrolled (
			    qw/ ROText -scrollbars e -wrap word relief sunken /
			    ) -> pack (
				       -fill => 'both',
				       -expand => 'yes',
				       -padx => 1,
				       -pady => 1,
				       );
    
    $text->configure (
		      -font => $config{ 'font' }
		      );

    $text->Tag (
		'norm',
		-font => $config{ 'font' }
		);

    $text->Tag (
		'bold',
		-font => $config{ 'font_bold' }
		);

    $text->Tag (
		'search',
		-font => $config{ 'font_bold' },
		-background => '#ce5555',
		-foreground => 'white'
		);


    $mw->bind (
	       '<Key>' => \&KeyPress
	       );

    if ( $OS ne 'WIN' ) {
	$SIG{ 'USR1' } = sub {

	    $mw->deiconify();
	    $mw->raise();
	    $entry->focus();
	};

	$SIG{ 'USR2' } = sub {
	    $mw->iconify();
	};
    }

    $mw->repeat (
		 100 => sub {}
		 );

    $list->bind (
		 '<Key>' => \&ListKey
		 );

    $list->bind (
		 '<<ListboxSelect>>' => sub { ListSel ( ($list->curselection)[0] ) }
		 );

    $list->activate ( 0 );
    $list->focus();

    $mw->geometry ($geometry) if ($geometry);
}


sub TextHighlight {
    my ( $t, $subs, $type ) = @_[0..2];

    printd 'TextHighlight';
    return unless length ( $subs );
    printd 'Highlighting...';
    $t->tagRemove( 'search', qw/ 0.0 end / );
    my( $current, $length ) = ( '1.0', 0 );

    while (1) {
	$current = $t->search ( -count => \$length, "-$type", $subs, $current, 'end' );
	last if not $current;
	$t->tagAdd ('search', $current, "$current + $length char" );
	$current = $t->index ("$current + $length char" );
    }
}


sub FindInArticles {
    my @pl = qw/ -side top -padx 10 -pady 1 -anchor w /;

    printd "FindInArticles";

    unless ( $sd->{ init } ) {
	printd "No dictionary loaded, return";
	return;
    }

    my $t = $mw->Toplevel();

    $t->title ( 'Search' );

    my $fr_t = $t ->Frame (
			  -relief => 'sunken',
			  -border => 1,
			  ) -> pack(
				   -side =>'top',
				   -fill => 'x',
				   -pady => 15,
				   -pady => 15,
				   );
    
    my $e = $fr_t->Entry (
			  -width => 50,
			  -font => $config{ 'font' },
			  -textvariable => \$search_entry
			  )-> pack (
				    -side =>'left',
				    -fill => 'x',
				    -expand => 'yes',
				    );

    $e->bind (
	      '<Key>' => [ \&SearchEntryPress ]
	      );

    $e->bind (
	      '<Escape>' => sub { $e -> delete(0.1, 'end'); }
	      );

    if ( $OS eq 'UNIX' && ! $config{ 'disable_switcher' }   ) {

	my $o = $fr_t->Optionmenu (
				   -options  => [ () ] ,
				   -variable => \$encode_convert_search,
				   -command  => [ \&ChangeEncodingSearch ]
				   ) -> pack(
					     -side =>'right'
					     );

	$o->addOptions (
			sort keys ( %encode_converts )
			);
    }

    my $fr_m1 = $t->Frame() -> pack (
				     -side =>'top',
				     -fill => 'x',
				     );
    
    my $fr_m1a = $fr_m1->Frame() -> pack (
					  -side =>'left',
					  -fill => 'x',
					  );

    my $fr_m1b = $fr_m1->Frame() -> pack (
					  -side =>'right',
					  -fill => 'x',
					  );

    my $lf_left = $fr_m1a->LabFrame ( -label => 'Start search from') -> pack ( @pl );

    for ( 'Dictionary start', 'Current position' ) {
	$rb1_value = $_ unless $rb1_value;
 	$lf_left->Radiobutton (
			       -text     => $_,
			       -value    => $_,
			       -variable => \$rb1_value,
			       -command  => \&SetSearchPosition
			       ) -> pack( @pl );
    }


    my $lf_right = $fr_m1b -> LabFrame ( -label => 'Search for') -> pack ( @pl );

    for ( 'Substring', 'Whole word', 'Perl regex' ) {
	$rb2_value = $_ unless $rb2_value;
	$lf_right -> Radiobutton (
				  -text     => $_,
				  -value    => $_,
				  -variable => \$rb2_value,
				  -command  => \&SetSearchType
				  )->pack ( @pl );
    }

    my $fr_b = $t->Frame() -> pack (
				    -side =>'bottom',
				    -fill => 'x'
				    );

    my $fr_b1 = $fr_b->Frame() -> pack (
					-side =>'top',
					-fill => 'x'
					);

    $fr_b1->Button (
		   -text        => 'Search',
		   -borderwidth => 1,
		   -command     => [ \&FindInArticlesFirst  ],
		   -font        => $config{ 'font_bold' },
		   ) -> pack (
			      -pady   => 10,
			      -padx   => 10,
			      -expand => 1,
			      -fill   => 'x',
			      -side   => 'left',
			      );
    $fr_b1->Button (
		   -text        => 'Search next',
		   -borderwidth => 1,
		   -command     => [ \&FindInArticlesNext  ],
		   -font        => $config{ 'font_bold' },
		   ) -> pack (
			      -pady   => 10,
			      -padx   => 10,
			      -expand => 1,
			      -fill   => 'x',
			      -side   => 'left',
			      );

    $fr_b1->Button (
		    -text        => 'Close',
		    -borderwidth => 1,
		    -command     => [ \&CloseFindInArticles, $t  ],
		    -font        => $config{ 'font_bold' },
		    ) -> pack (
			       -pady   => 10,
			       -padx   => 10,
			       -expand => 1,
			       -fill   => 'x',
			       -side   => 'right',
			       );

    my $fr_b2 = $fr_b->Frame() -> pack (
					-side =>'bottom',
					-fill => 'x'
					);


    $slab = $fr_b2->Label (
			    -text => q { },
			    -font => $config{ 'font' },
			    -justify => 'left' ) -> pack (
							 -side => 'left', 
							  );


    $fr_b2->ProgressBar (
			 -borderwidth => 2,
			 -relief      => 'sunken',
			 -width       => 20,
			 -padx        => 2,
			 -pady        => 2,
			 -variable    => \$search_status,
			 -resolution  => 0,
			 -from        => 0,
			 -blocks      => 100,
			 -colors      => [ 0 => 'blue' ],
			 -to          => 100
			 )->pack (
				  -padx   => 5,
				  -pady   => 10,
				  -side   => 'bottom',
				  -fill   => 'both',
				  -expand => 1
				  );

    $search_status = int  ( ( $curWordCnt * 100.0 ) / $sd->{ header }->{ words_total } );
    $stop_search = 0;

    $encode_convert_search =
	$encode_autoswitch{ $sd->{ header }->{ a_lang } }
    			if (
			    exists ( $encode_autoswitch{ $sd->{ header }->{ a_lang } } )
			    );
    
    CenterWindow ( $t );
}


sub CenterWindow {
    my $w = shift;
    return unless $w;
    $mw->update;
    $_ = $mw->geometry;
    @_ = split /[x\+]/, $_;
    my ( $m_w, $m_h, $m_x, $m_y ) = @_;
    $_ = $w->geometry;
    @_ = split /[x\+]/, $_;
    my ( $t_x, $t_y ) = @_[ 0,1 ];
    my $geom = $t_x . 'x' . $t_y . '+' .
			int ( ( $m_w  - $t_x ) / 2 + $m_x ) .
			    '+' . int ( ( $m_h - $t_y ) / 2 + $m_y ) ;

    $w->geometry( $geom );
}


sub CloseFindInArticles {
    my $t = shift;
    $t->destroy;
    $stop_search = 1;
}


sub SetSearchPosition {
    printd 'SetSearchPosition';
}


sub SetSearchType {
    printd 'SetSearchType';
}


sub FindInArticlesFirst {
    printd 'FindInArticlesFirst';

    printd "rb1_value = '$rb1_value'";
    printd "rb2_value = '$rb2_value'";


    return unless ( length ( $search_entry ) );

    $curWordCnt = 0;
    $search_status = 0;

    $searchSavePos = $sd->{ f_index_pos_cur };

    if ( $rb1_value eq 'Dictionary start') {
	$searchCurPos = $sd->{ f_index_pos }
    }
    else {
	$searchCurPos = $sd->{ f_index_pos_cur }; 
    }

    $sd->{ f_index_pos_cur } = $searchCurPos;

    DoDictionarySearch();

    printd 'restoring f_index_pos_cur';
    $sd->{ f_index_pos_cur } = $searchSavePos;
}


sub FindInArticlesNext {
    printd 'FindInArticlesNext';

    printd "rb1_value = '$rb1_value'";
    printd "rb2_value = '$rb2_value'";

    return unless ( length( $search_entry ) );

    return unless ( $searchCurPos );

    $searchSavePos = $sd->{ f_index_pos_cur };

    $sd->{ f_index_pos_cur } = $searchCurPos;

    DoDictionarySearch();

    printd 'restoring f_index_pos_cur';
    $sd->{ f_index_pos_cur } = $searchSavePos;
}

my $sls = '|';

sub UpdateProgressSearch {

    if ($sls eq '|') {
	$sls = '/';
    }
    elsif ($sls eq '/') {
	$sls = '-';
    }
    elsif ($sls eq '-') {
	$sls = '\\';
    } else {
	$sls = '|';
    }

    $slab->configure (
		      -text => $sls,
		      );
}

sub StopProgressSearch {
    $slab->configure (
		      -text => ' ',
		      );
}



sub DoDictionarySearch {
    printd 'DoDictionarySearch';
    $sls = '|';
    my $ct = 0;
    while (1) {
	if ($ct++ == 50) {
	    $ct = 0;
	    UpdateProgressSearch();
	}
	if ( $stop_search ) {
	    StopProgressSearch();
	    return;
	}
	my $nw = $sd->get_next_word;
	last if ($nw eq q{} );

	$curWordCnt++;

	$search_status = int ( ( $curWordCnt * 100.0 ) / $sd->{ header }->{ words_total } );
	$mw->update;

	# printd "search_status: '$search_status'";

	$nw = decode ("utf8", $nw );

	$searchCurPos = $sd->{ f_index_pos_cur }; 

	my $art = $sd->read_unit( $sd->{ cur_word_pos } + $sd->{ articles_pos } );
	$art = decode ( "utf8", $art ); 

	if ( $rb2_value eq 'Substring' ) {
	    if (
		( $nw  =~ /\Q$search_entry/ ) ||
		( $art =~ /\Q$search_entry/ )
	       ) {
		$nw = encode ( "utf8", $nw );
		PrintArticle ( $nw, $sd->{ cur_word_pos } );
		TextHighlight( $text, $search_entry, 'exact' );
		last;
	    }
	}
	elsif ( $rb2_value eq 'Whole word' ) {
	    if (
		( $nw eq $search_entry ) ||
		( $art =~ /\Q $search_entry / )
	       ) {
		$nw = encode ( "utf8", $nw );
		PrintArticle ( $nw, $sd->{ cur_word_pos } );
		TextHighlight ( $text, $search_entry, 'exact' );
		last;
	    }
	}
	elsif ($rb2_value eq 'Perl regex' ) {
	    if (
		( $nw =~ /$search_entry/ ) ||
		( $art =~ /$search_entry/ )
	       ) {
		$nw = encode ( "utf8", $nw );
		PrintArticle ( $nw, $sd->{ cur_word_pos } );
		TextHighlight( $text, $search_entry, 'regexp' );
		last;
	    }
	}
    }

    StopProgressSearch();
}


sub SearchEntryPress {
    my $s_enc;
    my $i=0;

    return unless ( length ( $search_entry ) );

    printd "SearchEntryPress()";

    my $s = substr ( $search_entry, 0, length ($search_entry ) - 1    );
    my $n = substr ( $search_entry,    length ($search_entry ) - 1, 1 );

    if ( $OS eq 'UNIX' && ! $config{ 'disable_switcher' } )
    {
	if ( unpack ("C", $n) >= 128 ) {
	    $s_enc = $encode_converts{ $encode_convert_search };
	    
	    if ( defined ( $s_enc ) && ( $s_enc ne q{} ) && ( $s_enc ne 'none' ) ) {
		from_to ( $n, $s_enc, "utf8" );
	    }
	    
	    $n = decode( "utf8", $n ) unless( utf8::is_utf8 ( $n ) );
	    $search_entry = $s . $n;
	}
    }
}


sub SafeListKeyUp {
    my $sel = ( $list->curselection() )[0];

    if ( $sel ) {
	$list->selectionClear ( 0, "end" );
	$list->selectionSet ( $sel -1 );
	$list->see ( $sel -1 );
    }
    ListKeyUp();
}


sub SafeListKeyDown {
    my $sel = ( $list->curselection() )[0];

    if ($sel < SDICT_LOAD_ITEMS - 1 ) {
	$list->selectionClear ( 0, "end" );
	$list->selectionSet ( $sel + 1 );
	$list->see ( $sel + 1 );
    }

    ListKeyDown();
}


sub ScrollCmd {
    my ( $f, $s ) = @_[0,1];
    my $sel = ( $list->curselection() )[0];

    printd "ScrollCmd()";

    CleanCacheHistory();

    if ( $f eq 'scroll' ) {
	if ($s == -1 ) {
	    printd "arrow up";
#	    if ($sel) {
#		$list->selectionClear(0, "end");
#		$list->selectionSet($sel -1);
#		$list->see($sel -1);
#	    }
#	    ListKeyUp();
	    #CleanCacheHistory();
	    SafeListKeyUp();

	}
	elsif ( $s == 1 ) {
	    printd "arrow down";

#	    if ($sel < SDICT_LOAD_ITEMS - 1) {
#		$list->selectionClear(0, "end");
#		$list->selectionSet($sel + 1);
#		$list->see($sel + 1);
#	    }
#	    ListKeyDown();
	    #CleanCacheHistory();
	    SafeListKeyDown();

	}
	else {
	    printd "Unknown param '$s'";
	}
	return;

    }
    elsif ( $f eq 'moveto' ) {

	my $ele = int ( $#{ $sd->{ sindex_3 } } * $s );
	$ele = 0 if ( $ele < 0 );
	
	my $l =  scalar ( @{ $sd->{ sindex_3 } } ) ;

	printd "going to $ele ($s) (max3: $l)";

	if ( $ele >= $l ) {
	    $ele = $l - 5;
	    printd "too big, set to $ele";
	}

	my $www = @{ $sd->{ sindex_3 } } [ $ele ];
	my ( $w, $p ) = @{ $www };

	LoadListFromPos( $sd->{ f_index_pos } + $p );
	ListSelEntryIns();
    }
}


sub ListKeyUp {
   my $sel = ( $list->curselection() )[0];

   printd "ListKeyUp(), sel=$sel";

   CleanCacheHistory();

   if ( $sel == 0 ) {

       $sd->{ f_index_pos_cur } = $curBot;
       $sd->get_next_word;
       $curBot = $sd->{ f_index_pos_cur };

       $sd->{ f_index_pos_cur } = $curUp;
       my $ww = $sd->get_prev_word;
       my $www = decode ( "utf8", $ww );

       if ( $ww ne q{} ) { 
	   $curUp = $sd->{ f_index_pos_cur };
	   $list->delete ( SDICT_LOAD_ITEMS - 1 );
	   $list->insert ( 0, $www );

	   pop ( @curListView );
	   unshift @curListView, [ $ww, $sd->{ cur_word_pos } ];
       }
   }

   ListSelEntryIns();
}


sub ListKeyDown {
    my $sel = ( $list->curselection() )[0];

    printd "ListKeyDown()";

    CleanCacheHistory();

    if ( $sel == SDICT_LOAD_ITEMS - 1 ) {

	$sd->{ f_index_pos_cur } = $curUp;
	$sd->get_next_word;
	$curUp = $sd->{ f_index_pos_cur };

	$sd->{ f_index_pos_cur } = $curBot;
	my $ww = $sd->get_next_word;
	my $www = decode ( "utf8", $ww );

	if ( $ww ne q{} ) { 

	    $curBot = $sd->{ f_index_pos_cur };

	    $list->delete( 0 );
	    $list->insert( 'end', $www );

	    shift ( @curListView );
	    push @curListView, [ $ww, $sd->{cur_word_pos} ];
	}
    }
    ListSelEntryIns();
}


sub ListKey {
    my( $widget ) = @_;
    my $e = $widget->XEvent;    # get event object
    my( $keysym_text, $keysym_decimal ) = ( $e->K, $e->N );

    printd "Key>$keysym_text<";

    if ( $keysym_text =~ /Down/i ) {
	ListKeyDown();
    }
    elsif  ( $keysym_text =~ /Up/i ) {
	ListKeyUp();
    }
}


sub ListSel {
    my $sel = shift;

    printd "ListSel(), sel = '$sel'";

    CleanCacheHistory();

    my ( $word, $pos )  = @{ $curListView [ $sel ] };
    printd "W> '$word'  P> '$pos'";
    PrintArticle ( $word, $pos );
    $list->focus;
}


sub ListSelEntryIns {
    printd 'ListSelEntryIns()';

    CleanCacheHistory();

    my $sel = ( $list->curselection() )[0];
    my ($word, $pos) = @{ $curListView [ $sel ] };
    # printd "W> '$word'  P> '$pos'";
    $word = decode ( "utf8", $word );
    $entry->delete ( 0, 'end' );
    $entry->insert ( 'end', $word );
}


sub ListSelHist {
    my ( $num, $list ) = @_;
    my $word = $HISTORY [ $num ];
    printd "ListSelHist, $word";
    return unless $word;
    $word = encode ( "utf8", $word );
    FindWord ( $word );
    $list->focus;
}


sub WordHistory {
    printd "WordHistory()";

    # return unless (@HISTORY);
    return if ( $toplevel_hist );

    $toplevel_hist = $mw->Toplevel();
    $toplevel_hist->title ( 'History' );

    my $fr = $toplevel_hist -> Frame() -> pack (
						-side   => 'top',
						-fill   => 'both',
						-expand => 'yes',
						);

    $list_h = $fr->Scrolled (
			     qw / Listbox -setgrid 1 -height 15 -scrollbars e /

			     ) -> pack (
					qw / -side left -expand yes -fill both /
					);
    $list_h->focus;
    $list_h->delete ( 0, 'end' );
    $list_h->insert ( 0, @HISTORY );
    $list_h->activate ( 0 );

    $list_h->bind (
		   '<<ListboxSelect>>' => sub {
		       	ListSelHist ( ( $list_h->curselection )[0], $list_h )
		       }
		   );

    my $fr2 = $toplevel_hist-> Frame () -> pack (
						 -side => 'bottom',
						 -fill => 'x'
						 );

    $fr2->Button (
		  -text        => 'Close',
		  -borderwidth => 1,
		  -command     => [ \&CloseHistory ],
		  -padx        => 30,
		  -pady        => 10,
		  -font        => $config{ 'font_bold' },
		  -underline   => 0
		  ) -> pack (
			     -side => 'bottom'
			     );

    CenterWindow( $toplevel_hist );
}


sub SaveArticleAsText {
      printd "SaveArticleAsText()";

      my $file_export = $mw->getSaveFile();

      return if (!defined($file_export)||($file_export eq q{}));

      printd "Selected file : '$file_export'\n";

      unless (open (OF, "> $file_export")) {
	  DisplayError("Cannot export file: $!");
	  return ;
      }

      my $t = $text->get('1.0', 'end');

      $t = encode("utf8", $t);

      if ( $config{ 'output_encoding' } )
      {
	  from_to ( $t, "utf8", $config{ 'output_encoding' } );
      }
      elsif ( ! $config{ 'disable_switcher' } && $encode_converts { $encode_convert } ne 'none' )
      {
	  from_to ( $t, "utf8",  $encode_converts { $encode_convert } );
      }

      print OF $t;
      close OF;

      printd "done";
}


sub DisplayError {
    my $text = shift;

    $text = 'Error: ' . $text;

    my $dialog = $mw->Dialog( -text => $text,
			      -font => $config{'font_bold'},
			      -bitmap => 'error',
			      -title => 'Error',
			      -default_button => 'Continue',
			      -buttons => [qw/Continue/] );
    my $d = $dialog->Show();
}


sub CloseHistory {
    $toplevel_hist->destroy;
    $toplevel_hist = undef;
    $list_h = undef;
}


sub UpdateHistoryList {
    return unless ( $list_h );
    $list_h->delete ( 0, 'end' );
    $list_h->insert ( 0, @HISTORY );
}


sub UpdateHistory ($) {
    my $word = shift;
    my $upd;

    printd "UpdateHistory; word '$word'";
    return unless $word;

    $upd = 1;
    for ( @HISTORY ) {
	if ( $word eq $_ ) {
	    printd "No need to update history";
	    $upd = 0;
	}
    }

    if ( $upd ) {
	printd "Update HISTORY";
	unshift @HISTORY, $word;
	my $l = SDICT_HISTORY_LEN - 1;
	@HISTORY = @HISTORY [ 0..$l ] if ( scalar ( @HISTORY ) > SDICT_HISTORY_LEN );
	UpdateHistoryList ();
    }

    return unless ( $history_back_upd );

    $upd = 1;
    for ( @HISTORY_BACK ) {
	if ( $word eq $_ ) {
	    printd "No need to update  history_back";
	    $upd = 0;
	}
    }

    if ( $upd ) {
	printd "Update HISTORY_BACK";
	push @HISTORY_BACK, $word;
	my $l = SDICT_HISTORY_BACK_LEN - 1;
	@HISTORY_BACK = @HISTORY_BACK [ 0..$l ] if ( scalar ( @HISTORY_BACK ) > SDICT_HISTORY_BACK_LEN );
    }

}


sub CleanHistory {
    @HISTORY = 
	@HISTORY_BACK = ();
}


sub BackButton {
    my (
	$i,
	$word,
	$cword,
	$dummy
	);

    printd "BackButton";
    
    return unless ( @HISTORY_BACK );

    my $sel = ( $list->curselection() )[0];
    if ( $sel ) {
	( $cword, $dummy ) = @{ $curListView [ $sel ] };
	$cword = encode ( "utf8", $cword );    
    }

    unless ( $cword ) {
	$word = pop ( @HISTORY_BACK );
    }
    else {
	for ( $i=0; $i < $#HISTORY_BACK; $i++ ) {
	    $word = pop ( @HISTORY_BACK );
	    printd "cword = $cword ;  word = $word";
	    last if ( ( $word eq $cword ) || ! ( $word ) );
	    $word = pop ( @HISTORY_BACK );
	}
    }

    return unless ( $word );

    $history_back_upd = 0;

    printd "Actual search...";
    $word = encode( "utf8", $word );    

    FindWord( $word );
    $history_back_upd = 1;
}


sub PrintArticle ( $$ ) {
    my ( $word, $pos ) = @_[ 0, 1 ];

    my (
	$forms,
	$trans
	);

    printd "print article; word '$word', pos '$pos'";

    my $art = $sd->read_unit ( $pos + $sd->{ articles_pos } );

    my $art_u = decode ( "utf8", $art ) ;

    $word = decode ( "utf8", $word );
    UpdateHistory ( $word );

    $entry->delete ( 0, 'end' );
    $entry->insert ( 'end', $word );

    $text->delete( "1.0", 'end' );

    $trans = $art_u;

    if ( $trans =~ m|<t>.+</t>| ) {
	$trans =~ s|.*<t>(.*?)</t>.*|$1|s;
    }
    else {
	$trans = q{};
    }

    $forms = $art_u;

    if ( $forms =~ m|<f>.+</f>| ) { 
	$forms =~ s|.*<f>(.*?)</f>.*|$1|s;
    }
    else {
	$forms = q{};
    }
#
# Print word
#
    $text->insert ( 'end', "$word",  ['bold'] );

#
# trans
#
    $text->insert ( 'end', " [ $trans ]", [ 'norm' ] ) if ( $trans );

#
# forms
#
#    printf ">>>$forms<<<\n";

    $text->insert ( 'end', " ( $forms )", [ 'norm' ] ) if ( $forms );

#
# Separator
#
    $text->insert ( 'end', "\n\n", [ 'norm' ] );

#
# TODO! Handle all tags
#
    $art_u =~ s|<br>|\n|isg;
    $art_u =~ s|<p>|\n\n|isg;

    $art_u =~ s|<l>||isg;
    $art_u =~ s|</l>||isg;
    $art_u =~ s|<li>||isg;

    $art_u =~ s|<b>||isg;
    $art_u =~ s|</b>||isg;

    $art_u =~ s|<i>||isg;
    $art_u =~ s|</i>||isg;

    $art_u =~ s|<u>||isg;
    $art_u =~ s|</u>||isg;

    $art_u =~ s|<t>.*?</t>||s;
    $art_u =~ s|<f>.*?</f>||s;

    $art_u =~ s|<sup>||s;
    $art_u =~ s|</sup>||s;


# AI format
    if ( $config{ 'format_article' } ) {
	$art_u =~ s|(\d{1,3}[\.\)])|\n\n$1|g;
	$art_u =~ s|\s([A-Za-z]\))|\n\n$1|g;

	$art_u =~ s|\s+([\x{410}-\x{44f}]\))|\n   $1|g;
	# $art_u =~ s|;\s+|\n\n\t|g;
    }

    $art_u =~ s|\r||isg;
    $art_u =~ s|^\s+||i;
    $art_u =~ s|^\n+||isg;
    $art_u =~ s|\n{2,}|\n\n|isg;

#    print ">>$art_u<<\n";

# /AI

#
#
# ?
#    $art_u =~ s|<.+?>||s;

    if ( $art_u =~ m|<r>.+?</r>| ) { # article with links

	my $tag = 1;
	my $tag_s = q{};
	my %tg = ();

	$text->tag ( qw / configure color1 -background/ => '#a0b7ce' );

	while ( $art_u =~ s|(.+?)<r>(.+?)</r>||is ) {

	    my ( $normal, $linked ) = ( $1, $2 );

	    #printd "before: '$normal'";
	    #printd "link: '$linked'";

	    $text->insert ( 'end', "$normal ", [ 'norm' ] );

            $tag_s = 'd' . $tag;
            $tag++;

            $text->insert ( 'end', $linked, [ $tag_s, 'color1' ] );

            $tg{ $tag_s } = $linked;
	}

	$text->insert ( 'end', $art_u, [ 'norm' ] );

	for my $tgg ( keys ( %tg ) ) {
	    $text->tagBind ( $tgg, '<Any-Enter>' => sub { shift->tagConfigure ( $tgg, @bold   ) } );
	    $text->tagBind ( $tgg, '<Any-Leave>' => sub { shift->tagConfigure ( $tgg, @normal ) } );
	    my @aa = ( $tgg, "<1>" );
	    my $wrd = $tg{ $tgg };
	    $text->tagBind ( @aa  => sub { &FindWord ( $wrd ) } );
	}


    }
    else { # article with no links

	$text->insert ( 'end', $art_u, [ 'norm' ] );
    }
}


sub FindWord {
    my $word = shift;
    my $word_u = decode ( "utf8", $word );
    my $ref;
    my $search_pos = -1;

    printd "Searching word '$word'";

    my $len = length( $word_u );
    my $subw = substr ($word_u, 0, 3 );
    printd "len '$len'";

    return unless $len;


    for (my $i=1; $i<4; $i++ ) {

	if ( $i == 1 ) {
	    $ref = $sd->{ sindex_1 };
	}
	elsif ( $i == 2 ) {
	    $ref = $sd->{ sindex_2 };
	}
	else {
	    $ref = $sd->{ sindex_3 };
	}

	printd "Searching '$i'";

	for my $j ( @$ref ) {
	    # printd "$_";
	    my ($wo, $ndx) = @$j;
	    # $wo = decode ("utf8", $wo);
	    # printd "wo '$wo' $ndx";

	    if ( substr ( $wo, 0, $i ) eq substr ( $subw, 0, $i ) ) {
		printd "Found in '$i', wo: '$wo', ndx: '$ndx'";
		$search_pos = $ndx;
		# $i++;
		next;
	    }
	}
    }

    if ( $search_pos < 0 ) {
	printd "Not found";
	return;
    }

    printd "Scanning from pos '$search_pos'";

    my $findes_saved = $sd->{ f_index_pos_cur };

    $sd->{ f_index_pos_cur } = $search_pos + $sd->{ f_index_pos };

    for ( my $ii=0; $ii < Sdict::SDICT_SEARCH_FORWARD; $ii++ ) {
	my $prev_pos = $sd->{ f_index_pos_cur };
	my $nw = $sd->get_next_word;

	if ( $nw eq q{} ) {
	    $sd->{ f_index_pos_cur } = $findes_saved;
	    printd 'Not found';
	    return;
	}

	$nw = decode ( "utf8", $nw );

	if ( $word_u eq $nw ) {
	    printd "!!! found";
	    LoadListFromPos ( $prev_pos );
	    ReturnPress();

	    $list->focus;
	    return;
	}
    }

    $sd->{ f_index_pos_cur } = $findes_saved;
    printd 'Not found';
}


sub EntryPaste {
    printd "paste, entry '$entry_text'";
    $entry_text = decode ( "utf8", $entry_text );
    EntryPressAction ( 1, q{} );

    my $word = encode ( "utf8", $entry_text );
    FindWord( $word );
}


sub EntryPress {
    # my $s_enc;
    # my $i=0;
    printd "EntryPress()";

    my $e = $entry->XEvent;
    my( $keysym_text, $keysym_decimal) = ( $e->K, $e->N );

    printd "Key '$keysym_text'";

    if ( $keysym_text =~ /^Up$/i ) {
	printd 'Up in the entry';
	SafeListKeyUp();
	return;
    }

    if ( $keysym_text =~ /^Down$/i ) {
	printd 'Down in the entry';
	SafeListKeyDown();
	return;
    }

    EntryPressAction ( 0, $keysym_text );
}


sub CleanCacheHistory {
    printd 'CleanCacheHistory()';
    @HIT_CACHE = ();
}


sub IsCacheHistoryValid {
    my $val;

    # for (my $i=0; $i< $#HIT_CACHE; $i++) { printd "$i >$HIT_CACHE[$i]<"; }

    return 0 unless ( $config{ 'cache_search_results' } );

    if ( defined $HIT_CACHE [1] ) {
	$val = 1;
    }
    else {
	$val =0;
    }

    printd "IsCacheHistoryValid = $val";
    return $val;
}


sub EntryPressAction {
    my ( $do_load, $key ) = @_;
    my $s_enc;
    my $i=0;

    return unless ( length ( $entry_text ) );

    printd "entry '$entry_text'";

    my $s = substr ( $entry_text, 0, length($entry_text) - 1    );
    my $n = substr ( $entry_text,    length($entry_text) - 1, 1 );


    if ( $key !~ /BackSpace/i ) { 
	if ( $OS eq 'UNIX' && ! $config{ 'disable_switcher' } )
	{
	    if ( unpack ( "C", $n ) >= 128 ) {
		$s_enc = $encode_converts{ $encode_convert };

		if ( defined ( $s_enc ) && ( $s_enc ne q{} ) && ( $s_enc ne 'none' ) ) {
		    from_to ( $n, $s_enc, "utf8" );
		}

		$n = decode ( "utf8", $n ) unless ( utf8::is_utf8 ( $n ) );
		$entry_text = $s . $n;
	    }
	}
    }

    if ( $adaptive_keyb ) {
	my $elapsed = tv_interval ( $t0, [ gettimeofday ] );
	$t0 = [ gettimeofday ];

	if ( $elapsed < SDICT_KEYPRESS_SENS ) {
	    printd "only '$elapsed' between keys";
	    return;
	}
    }

    my $elen = length ( $entry_text );
    printd "LEN '$elen'";

    my $pl = $prev_len;
    $prev_len = $elen;

    my $drct = 0;
    if ( $elen >= $pl ) {
	$drct = 1;
	printd '>>> forward';
    }
    else {
	$drct = 0;
	printd '<<< backward';
    }

    if ( $elen >3 && IsCacheHistoryValid() ) { # only if need seek, not only sindex

	if ( $drct ) { # >>>
	    unless ( $HIT_CACHE [ $elen-1 ] ) {
		printd 'FORWARD: prev was failed, return';
		return;
	    } 
	}
	else { # <<<
	    if ( $HIT_CACHE [ $elen ] ) {
		printd 'BACKWARD: cache hit found';
		LoadListFromPos ( $HIT_CACHE [ $elen ] );
		return;
	    }
	    else {
		printd 'BACKWARD: no cache hit, return';
		return;
	    }
	}
    }

    $HIT_CACHE [ $elen ] = 0;

    if ( $elen == 1 ) {
	for ( @{ $sd->{ sindex_1 } } ) {
	    my ( $w, $p ) = @{ $_ };
	    if ( $w eq $entry_text ) {
		$HIT_CACHE [ $elen ] = $sd->{ f_index_pos } + $p;
		LoadListFromPos ( $sd->{ f_index_pos } + $p );
		# my $tot = $#{$sd->{sindex_1}};
		# my $moveto = ($i * SDICT_SCROLL_MADIC) / $tot ;

		my $procent = $i / $#{ $sd->{ sindex_1 } };

		$lscroll->set ( $procent, $procent + SDICT_SCROLL_MAGIC );

		return;
	    }

	    $i++;
	}
    }
    elsif ( $elen == 2 ) {
	for ( @{ $sd->{ sindex_2 } } ) {
	    my ( $w, $p ) = @{ $_ };
	    if ( $w eq $entry_text ) {
		$HIT_CACHE [ $elen ] = $sd->{ f_index_pos } + $p;
		LoadListFromPos ( $sd->{ f_index_pos } + $p );
		return;
	    }
	}
    }
    elsif ( $elen == 3 ) {
	for ( @{ $sd->{ sindex_3 } } ) {
	    my ( $w, $p ) = @{ $_ };
	    if ( $w eq $entry_text ) {
		$HIT_CACHE [ $elen ] = $sd->{ f_index_pos } + $p;
		LoadListFromPos( $sd->{ f_index_pos } + $p );
		return;
	    }
	}
    }
    else {
	printd "* >3";
	my $subw = substr ( $entry_text, 0, 3 );

	printd "for '$subw'";

	my $i = -1;

	for ( @{ $sd->{ sindex_3 } } ) {
	    $i++;
	    my ( $w, $p ) = @{ $_ };

	    if ( $w eq $subw ) {
		LoadListFromPos ( $sd->{ f_index_pos } + $p ) if ( $do_load );

		# search
		printd "Searching...";
		$sd->{ f_index_pos_cur } = $curUp;

		for ( my $ii=0; $ii < Sdict::SDICT_SEARCH_FORWARD; $ii++ ) {
		    my $prev_pos = $sd->{ f_index_pos_cur };
		    my $nw = $sd->get_next_word;

		    return if ( $nw eq q{} );

		    $nw = decode ( "utf8", $nw );

                    if ( ( !$ignoresindex ) &&
			 ( substr ( $entry_text, 0, 3 ) ne substr ( $nw, 0, 3 ) )
			 ) {
                        printd "another index entry entry '$entry_text', next word '$nw'";
                        return;
		    }


		    if ( $entry_text eq substr ( $nw, 0, length ( $entry_text ) ) ) {
			printd "nw '$nw'";
			$HIT_CACHE [ $elen ] = $prev_pos;
			LoadListFromPos ( $prev_pos );
			return;
		    }
		}
		return;
	    }
	}
    }
}


sub LoadListFromPos {
    my $pos = shift;
    $sd->{ f_index_pos_cur } = $pos;
    $curUp = $pos;
    my $cwu;
    @curListView = ();

    $list->delete ( 0, 'end' );

    for ( my $ii=0; $ii < SDICT_LOAD_ITEMS; $ii++ ) {

	$curWord = $sd->get_next_word;
	$cwu = decode ( "utf8", $curWord );
	last if ($curWord eq q{} );

	$list->insert ( 'end', $cwu );

	push @curListView, [ $curWord, $sd->{ cur_word_pos } ];
    }

    # my $curUp = $sd->{f_index_pos};
    $curBot = $sd->{ f_index_pos_cur };

    # $list->activate(0);

    $list->selectionClear ( 0, "end" );

    $list->see ( 0 );
    $list->selectionSet ( 0 );
    # $list->focus();

}


sub ReturnPress {
    printd "Return";
    my $sel = ( $list->curselection ) [0] ;

    CleanCacheHistory();

    # print_art (@{$sd->{words_list}}[$sel[0]]) if (@sel);

    my ( $word, $pos ) = @ { $curListView [ $sel ] };

    printd "W> '$word'  P> '$pos'";

    PrintArticle ( $word, $pos );
}


sub KeyPress {
    my( $widget ) = @_;
    my $e = $widget->XEvent;    # get event object
    my( $keysym_text, $keysym_decimal ) = ( $e->K, $e->N );

    printd "pressed '$keysym_text'";

    if ( ( $keysym_text =~ /^z$/i ) && ( $ControlPressed ) ) {
	$ControlPressed = 0;
	printd "Ctrl-Z";
	$mw->iconify();
	return;
    }

    if ( ( $keysym_text =~ /^\d$/i ) && ( $AltPressed ) ) {
	$AltPressed = 0;
	my $altstroke = "Alt-$keysym_text";
	my $altstrokedic = $config{ alt_bindings }->{ $altstroke };
	printd "pressed $altstroke --> $altstrokedic";

	if ( $DICTS { $altstrokedic } ) {
	    printd "loading $altstrokedic";
	    $dictionary = $altstrokedic;
	    ChangeDic();
	}
	else {
	    printd 'dictionary not found';
	}
	return;
    }

    if ( $keysym_text =~ /Control_/i ) {
	$ControlPressed = 1;
	return;
    }
    else {
	$ControlPressed = 0;
    }

    if ( $keysym_text =~ /Alt_/i ) {
	$AltPressed = 1;
	return;
    }
    else {
	$AltPressed = 0;
    }

}

sub PrintHelp {
    $fh->configure ( -state => 'disabled' );
    $toplevel = $mw->Toplevel ();
    $toplevel->title ( 'Help' );

    my $fr = $toplevel -> Frame (
				 -relief => 'sunken', #'solid', #'raised',
				 -borderwidth => 1,
				 ) -> pack (
					    -side =>'top',
					    -fill => 'x',
					    -padx => 5,
					    -pady => 5,
					    );

    my $text = <<EOS;

Perl/Tk Sdictionary. Version: $VERSION ($OS).

Written by Alexey Semenoff [http://swaj.net] (c) 2001-2007.

Perl: $]; Compress::Zlib: $Compress::Zlib::VERSION; Tk: $Tk::VERSION; Sdict: $Sdict::VERSION.

EOS

    $fr ->Label (
		 -image => $img
		 ) -> pack(
			   -side => 'left',
			   -padx => 15,
			   -pady => 15
			   );

    $fr->Label ( -text => $text,
		 -font => $config{ 'font_bold' },
		 -justify => 'center' ) -> pack (
						 -side   => 'top',
						 -expand => 'yes',
						 -padx   => 10,
						 -pady   => 10,
						 -fill   => 'both'
						 );

    my $fr2 = $toplevel->Frame (
				-borderwidth => 10,
				) -> pack (
					   -side => 'bottom',
					   -fill => 'x',
					   -expand => 'no',
					   );

    $fr2->Button (
		  -text => 'Ok',
		  -borderwidth => 1,
		  -command => [ \&ClosePrintHelp ],
		  -padx => 30,
		  -pady => 10,
		  -font => $config{ 'font_bold' },
		  -underline => 0
		  ) -> pack (
			     -side => 'bottom'
			     );

    CenterWindow ( $toplevel );
}


sub ClosePrintHelp {
    $toplevel->destroy;
    $fh->configure ( -state => 'normal' );
}


sub FileOpen {
    my @types = (
		 [ "Sdict files", [qw/.dct/] ],
		 [ "All files"  , '*'        ]
		);

    my $file_ = $mw->getOpenFile (
				  -filetypes  => \@types,
				  -initialdir =>  $config{ dictionary_dir }
				  );

    return if ( !defined ( $file_ ) || ( $file_ eq q{} ) );

    print "Selected dictionary: '$file_'\n";

    $sd->init (
	       { file => $file_ }
	       );

    unless ( $sd->load_dictionary_fast ) {
	prerror "Unable load dictionary";
	ChangeDic();
	return;
    }

    printd $sd->{ header }->{ title };

    if ( exists( $DICTS { $sd->{ header }->{ title } } ) ) {
	printd 'already in the list';
    }
    else {
	printd 'adding new dicttionary';

	$DICTS { $sd->{ header }->{title } } = $file_;

	$opt->addOptions ( $sd->{ header }->{ title } );
    }

    $dictionary = $sd->{ header }->{ title };

    $sd->init (
	       { file => $DICTS { $sd->{ header }->{ title } } }
	       );
    LoadDictionary();
}


sub DictionaryInfo {
    return if $toplevel_info_t;
    my $csp = 80;
    $toplevel_info_t = 1;
    $toplevel_info = $mw->Toplevel ();
    $toplevel_info->title ( 'Info' );

    my $fr = $toplevel_info->LabFrame (
				       -label => 'Dictionary',
				       ) ->pack (
						 -side =>'top',
						 -fill => 'both',
						 -expand => 'yes',
						 -padx => 10,
						 -pady => 1,
						 -anchor => 'w',
					     );
    my $txt = q{};
    $txt .= q{ } x $csp; 

    if ($sd->{ init }) {
	my $slv = $sd->{ slevels };
	$txt .= "\nTitle: " .
	    $sd->{ header }->{ title } . '. ' .
	    'Version: ' . $sd->{ header }->{ version } . '. ' .
	    'Words: ' . $sd->{ header }->{ words_total } . " ($slv+1)\n"  .
	    'Copyright: ' . $sd->{ header }->{ copyright } ;
	
	$txt = decode ( "utf8", $txt );
    }
    else {
	$txt .= "\nNo dictionary loaded!";
    }

    $txt .= "\n";

    $fr->Label (
		-text    => $txt,
		-font    => $config{ 'font_bold' },
		-justify => 'left',
		) -> pack (
			   -side   => 'top',
			   -expand => 'yes',
			   -padx   => 10,
			   -pady   => 10,
			   -fill   => 'both',
			   );
    
    my $fr2 = $toplevel_info -> Frame() -> pack (
						 -side => 'bottom',
						 -fill => 'x',
						 -padx => 30,
						 -pady => 10,
						 );

    my $fr3 = $toplevel_info->LabFrame (
				       -label => 'Configuration',

				       ) ->pack (
						 -side =>'bottom',
						 -fill => 'both',
						 -expand => 'yes',
						 -padx => 10,
						 -pady => 1,
						 -anchor => 'w',

					     );
    my $txt2 = q{};
    $txt2 .= q{ } x $csp; 
    $txt2 .= "\nDictionary dir: $config{'dictionary_dir'}\n";
    $txt2 .= "Cache search results: $config{'cache_search_results'}\n";
    $txt2 .= "Remember last dictionary: $config{'store_last_dic'}\n";
    $txt2 .= "Format article: $config{'format_article'}\n";


    $fr3->Label (
		-text    => $txt2,
		-font    => $config{ 'font_bold' },
		-justify => 'left',
		) -> pack (
			   -side   => 'top',
			   -expand => 'yes',
			   -pady   => 10,
			   -padx   => 10,
			   -fill   => 'both',
			   );

    $fr2->Button (
		  -text        => 'Close',
		  -borderwidth => 1,
		  -command     => [ \&CloseDictionaryInfo ],
		  -padx        => 30,
		  -pady        => 10,
		  -font        => $config{ 'font_bold' },
		  -underline   => 0
		  ) -> pack (
			     -side => 'bottom'
			     );

    CenterWindow( $toplevel_info );
}


sub CloseDictionaryInfo {
    $toplevel_info->destroy if $toplevel_info_t;
    $toplevel_info_t = 0;
}


sub ChangeDic
{
    $sd->unload_dictionary;
    my $dicc = encode( "utf8", $dictionary );
    printd "loading '$dicc'";
    $sd->init ( { file => $DICTS{ $dicc } } );
    LoadDictionary();
}


sub LoadDictionary
{
    unless ( $sd->load_dictionary_fast ) {
	prerror "Unable to load dictionary from file '", $DICTS{ $dictionary }, "'";
    }

    $searchCurPos = 0;
    $curWordCnt = 0;
    $search_status = 0;

    $curUp = 0;
    $curBot = 0;
    LoadListFromPos( $sd->{ f_index_pos } );
    $lscroll->set ( 0, SDICT_SCROLL_MAGIC );
    $list->activate ( 0 );

    if ( $cleanFirstTime++ ) {
	$text->delete ( "1.0", 'end' );
    }

    my $fword = decode ( "utf8", $curListView [0] [0] ); #@curListView->[0]->[0]);

    $entry->delete ( 0, 'end' );
    $entry->insert ( 'end', $fword );

    $encode_convert =
	$encode_autoswitch{ $sd->{ header }->{ w_lang } } 
                             if ( exists ( $encode_autoswitch{ $sd->{ header }->{ w_lang } } ) );

    CleanHistory();
    CleanCacheHistory();
    UpdateHistoryList();
    UpdateStatusBar();

    prinfo "Loaded $dictionary";
}


sub UpdateStatusBar {
    prinfo "UpdateStatusBar()";

    my $ltext = q{};

    if ($sd->{ init } ) {
	$ltext .= $sd->{ header }->{ title } . ' (Ver ' .
	    $sd->{ header }->{ version } . ', ' .
		    $sd->{ header }->{ words_total } . ' words)';

	$ltext = decode ( "utf8", $ltext );
	$ltext = substr ($ltext, 0, 100);
    }
    else {
	$ltext .= 'No dictionary loaded';
    }

    $fr4_lab->configure (
			 -text => $ltext,
			 );
}


sub ChangeEncoding {
    printd "Use encoding (main) '$encode_convert'";
}


sub ChangeEncodingSearch {
    print "Use encoding (search) '$encode_convert_search'\n";
}


__DATA__
/* XPM */
static char *logo_sdict_32x32_16[] = {
/* columns rows colors chars-per-pixel */
"32 32 16 1",
"  c #391E0E",
". c #203B4D",
"X c #235170",
"o c #4D2E19",
"O c #684D32",
"+ c #64676A",
"@ c #356789",
"# c #5A7588",
"$ c #748897",
"% c #835F3C",
"& c #8F7355",
"* c #A28560",
"= c #93999E",
"- c #BBBEC0",
"; c #D1D2D3",
": c #FEFEFE",
/* pixels */
":::::::::::;==$$-=;:::::::--::::",
"::::::::;$XXXXXXXXXX@=;::-X$::::",
":::::::$XXXX#=-;--$@XXXXXXX$::::",
":::::;@@@X@;::::::::;@XXX.X$::::",
"::::;@@X@$::::::::::::-XX.X$::::",
"::::$@X@X::::::::::::::;X+X$::::",
":::;@@.X$:::::::::::::::-XX$::::",
":::-##.#X::::::::::::::::=X=::::",
":::-$@@++-:::::::::::::::;+$::::",
":::-@#@.#@-:::::::::::::::$=::::",
":::;$###X#@=;:::::::::::::;-::::",
"::::$$+##++##=::::::::::::::::::",
":::::$$++$$+###=;:::::::::::::::",
"::::::=$#++$$#+$$=;:::::::::::::",
":::::::-$=++$==##$==-:::::::::::",
":::::::::-==#++==##$==-:::::::::",
":::::::::::====++==$====;:::::::",
":::::::::::::-===++===&&O+;:::::",
":::::::::::::::;===+++OOO%&;::::",
":::::::::::::::::;&oOOO**%&&;:::",
":::::::::::::::::::;**&%**O&&;::",
":::::::::::::::::::::-**%**O&=::",
":::;&:::::::::::::::::;&%%%o%+::",
":::;%-:::::::::::::::::-%%%oOO::",
":::;&*::::::::::::::::::%%OoOo::",
":::;%%*:::::::::::::::::&OOoo+::",
":::;%OO;::::::::::::::::OOo o-::",
":::-%oOO;::::::::::::::-Oo oO:::",
":::-oo  o=::::::::::::-Oo  o::::",
":::-o     o=;:::::::;&oo  &:::::",
":::-oo+&Oo   o%&=&&ooo  +-::::::",
":::-o-::::;-&+o   o o%=;::::::::"
};
__END__
